# -*- coding: utf-8 -*-
import json as pyjson

from abc import ABC, abstractmethod
from datetime import datetime, timedelta
from decimal import Decimal
from pathlib import Path
from sys import modules as _sys_modules
from typing import Any, Callable, List, Optional, Tuple, Type, Union

try:
    import dataclasses
except ImportError:
    dataclasses = None  # type: ignore

try:
    import rapidjson

    RapidJSONEncoder = rapidjson.Encoder
    RapidJSONDecoder = rapidjson.Decoder
    RapidJSONDecodeError = rapidjson.JSONDecodeError

except ImportError:
    rapidjson = None
    RapidJSONEncoder = None
    RapidJSONDecoder = None
    RapidJSONDecodeError = None

try:
    import orjson

except ImportError:
    orjson = None  # type: ignore

try:
    import numpy as np
except ImportError:
    np = None

JSONLIB_DEFAULT_PREFERENCE = ['orjson', 'rapidjson']


def _json_encode_default(obj: Any) -> Any:
    if np:
        if isinstance(obj, np.floating):
            return float(obj)
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, (np.ndarray, np.generic)):
            return obj.tolist()
    if dataclasses:
        if dataclasses.is_dataclass(obj):
            return dataclasses.asdict(obj)
    if isinstance(obj, bytes):
        return str(obj, encoding='utf-8')
    if isinstance(obj, Path):
        return str(obj)
    if isinstance(obj, datetime):
        return obj.isoformat()
    if isinstance(obj, timedelta):
        return obj.total_seconds()
    if isinstance(obj, Decimal):
        return str(obj)
    if hasattr(obj, 'eject'):
        return obj.eject()
    if hasattr(obj, 'to_dict'):
        return obj.to_dict()
    if hasattr(obj, 'dict'):
        return obj.dict()

    raise TypeError('Cannot encode obj as JSON: {}'.format(str(obj)))


class JsonLibABC(ABC):
    lib = 'json'

    @staticmethod
    @abstractmethod
    def dumps(
        data: Any,
        pretty: bool = False,
        sort_keys: bool = False,
        append_newline: bool = False,
        default: Optional[Callable[[Any], Any]] = None,
        **kwargs: Any,
    ) -> str:
        ...

    @staticmethod
    @abstractmethod
    def dumpb(
        data: Any,
        pretty: bool = False,
        sort_keys: bool = False,
        append_newline: bool = False,
        default: Optional[Callable[[Any], Any]] = None,
        **kwargs: Any,
    ) -> bytes:
        ...

    @staticmethod
    @abstractmethod
    def loads(string: str, **kwargs: Any) -> Any:
        ...

    @staticmethod
    @abstractmethod
    def useable() -> bool:
        ...

    @staticmethod
    def default(obj: Any) -> Any:
        """Default encoder"""
        return _json_encode_default(obj)

    @staticmethod
    def has_rapidjson() -> bool:
        return rapidjson is not None

    @staticmethod
    def has_orjson() -> bool:
        return orjson is not None


class JSON_STDLIB(JsonLibABC):
    lib = 'json'
    JSONEncoder = pyjson.JSONEncoder
    JSONDecoder = pyjson.JSONDecoder
    JSONDecodeError = pyjson.JSONDecodeError

    @staticmethod
    def dumps(
        data: Any,
        pretty: bool = False,
        sort_keys: bool = False,
        append_newline: bool = False,
        default: Optional[Callable[[Any], Any]] = None,
        **kwargs: Any,
    ) -> str:
        separators = (',', ': ') if pretty else (',', ':')
        dump_str = pyjson.dumps(
            data,
            indent=2 if pretty else None,
            sort_keys=sort_keys,
            separators=separators,
            default=default or _json_encode_default,
            **kwargs,
        )
        if append_newline:
            return f'{dump_str}\n'
        return dump_str

    @staticmethod
    def dumpb(
        data: Any,
        pretty: bool = False,
        sort_keys: bool = False,
        append_newline: bool = False,
        default: Optional[Callable[[Any], Any]] = None,
        **kwargs: Any,
    ) -> bytes:
        return JSON_STDLIB.dumps(
            data,
            pretty=pretty,
            sort_keys=sort_keys,
            append_newline=append_newline,
            default=default or _json_encode_default,
            **kwargs,
        ).encode()

    @staticmethod
    def loads(string: str, **kwargs: Any) -> Any:
        return pyjson.loads(string, **kwargs)

    @staticmethod
    def useable() -> bool:
        return True


class ORJSON(JsonLibABC):
    lib = 'orjson'

    @staticmethod
    def dumps(
        data: Any,
        pretty: bool = False,
        sort_keys: bool = False,
        append_newline: bool = False,
        default: Optional[Callable[[Any], Any]] = None,
        **kwargs: Any,
    ) -> str:
        return ORJSON.dumpb(
            data,
            pretty=pretty,
            sort_keys=sort_keys,
            append_newline=append_newline,
            default=default or _json_encode_default,
        ).decode(encoding='utf-8')

    @staticmethod
    def dumpb(
        data: Any,
        pretty: bool = False,
        sort_keys: bool = False,
        append_newline: bool = False,
        default: Optional[Callable[[Any], Any]] = None,
        **kwargs: Any,
    ) -> bytes:
        option = 0
        if pretty:
            option |= orjson.OPT_INDENT_2
        if sort_keys:
            option |= orjson.OPT_SORT_KEYS
        if append_newline:
            option |= orjson.OPT_APPEND_NEWLINE
        if 'numpy' in _sys_modules:
            option |= orjson.OPT_SERIALIZE_NUMPY

        return orjson.dumps(
            data, option=option, default=default or _json_encode_default
        )

    @staticmethod
    def loads(string: str, **kwargs: Any) -> Any:
        return orjson.loads(string)

    @staticmethod
    def useable() -> bool:
        return JsonLibABC.has_orjson()


class RAPIDJSON(JsonLibABC):
    lib = 'rapidjson'

    JSONEncoder = RapidJSONEncoder
    JSONDecoder = RapidJSONDecoder
    JSONDecodeError = RapidJSONDecodeError

    @staticmethod
    def dumps(
        data: Any,
        pretty: bool = False,
        sort_keys: bool = False,
        append_newline: bool = False,
        default: Optional[Callable[[Any], Any]] = None,
        **kwargs: Any,
    ) -> str:
        dump_str = str(
            rapidjson.dumps(
                data,
                indent=2 if pretty else None,
                sort_keys=sort_keys,
                default=default or _json_encode_default,
                datetime_mode=rapidjson.DM_ISO8601,
                **kwargs,
            )
        )
        if append_newline:
            return f'{dump_str}\n'
        return dump_str

    @staticmethod
    def dumpb(
        data: Any,
        pretty: bool = False,
        sort_keys: bool = False,
        append_newline: bool = False,
        default: Optional[Callable[[Any], Any]] = None,
        **kwargs: Any,
    ) -> bytes:
        return RAPIDJSON.dumps(
            data,
            pretty=pretty,
            sort_keys=sort_keys,
            append_newline=append_newline,
            default=default or _json_encode_default,
            **kwargs,
        ).encode()

    @staticmethod
    def loads(string: str, **kwargs: Any) -> Any:
        return rapidjson.loads(string, **kwargs)

    @staticmethod
    def useable() -> bool:
        return JsonLibABC.has_rapidjson()


def pick_lib() -> 'Type[JsonLibABC]':
    if ORJSON.useable():
        return ORJSON
    if RAPIDJSON.useable():
        return RAPIDJSON
    return JSON_STDLIB


def _import_rapidjson() -> 'Type[RAPIDJSON]':
    if RAPIDJSON.useable():
        return RAPIDJSON
    raise ImportError('rapidjson (python-rapidjson) not installed')


def _import_orjson() -> 'Type[ORJSON]':
    if ORJSON.useable():
        return ORJSON
    raise ImportError('orjson not installed')


def _import_json_stdlib() -> 'Type[JSON_STDLIB]':
    return JSON_STDLIB


def import_json(
    jsonlibs: Optional[Union[Tuple[str], List[str]]] = None
) -> 'Type[JsonLibABC]':
    lib2funk = {
        'rapidjson': _import_rapidjson,
        'orjson': _import_orjson,
        'json': _import_json_stdlib,
    }
    jsonlibs = jsonlibs or JSONLIB_DEFAULT_PREFERENCE

    for mod in jsonlibs:
        try:
            return lib2funk[mod]()
        except (ImportError, ModuleNotFoundError):
            pass

    return _import_json_stdlib()


class JsonLib:
    _jsonlib: Type[JsonLibABC]

    def __init__(self, jsonlib: Optional[Type[JsonLibABC]] = None):
        if jsonlib:
            self._jsonlib = jsonlib
        else:
            self._jsonlib = import_json()

    def dumps(
        self,
        data: Any,
        pretty: bool = False,
        sort_keys: bool = False,
        append_newline: bool = False,
        default: Optional[Callable[[Any], Any]] = None,
        **kwargs: Any,
    ) -> str:
        return self._jsonlib.dumps(
            data=data,
            pretty=pretty,
            sort_keys=sort_keys,
            append_newline=append_newline,
            default=default,
            **kwargs,
        )

    def dumpb(
        self,
        data: Any,
        pretty: bool = False,
        sort_keys: bool = False,
        append_newline: bool = False,
        default: Optional[Callable[[Any], Any]] = None,
        **kwargs: Any,
    ) -> bytes:
        return self._jsonlib.dumpb(
            data=data,
            pretty=pretty,
            sort_keys=sort_keys,
            append_newline=append_newline,
            default=default,
            **kwargs,
        )

    def loads(self, string: str, **kwargs: Any) -> Any:
        return self._jsonlib.loads(string, **kwargs)

    def use_orjson(self) -> None:
        self._jsonlib = _import_orjson()

    def use_rapidjson(self) -> None:
        self._jsonlib = _import_rapidjson()

    def use_json_stdlib(self) -> None:
        self._jsonlib = _import_json_stdlib()

    def use_json(self) -> None:
        self.use_json_stdlib()

    def which(self) -> str:
        return self._jsonlib.lib


JSONLIB: Type[JsonLibABC] = import_json()


def dumps(
    data: Any,
    pretty: bool = False,
    sort_keys: bool = False,
    append_newline: bool = False,
    default: Optional[Callable[[Any], Any]] = None,
    **kwargs: Any,
) -> str:
    return JSONLIB.dumps(
        data=data,
        pretty=pretty,
        sort_keys=sort_keys,
        append_newline=append_newline,
        default=default,
        **kwargs,
    )


def dumpb(
    data: Any,
    pretty: bool = False,
    sort_keys: bool = False,
    append_newline: bool = False,
    default: Optional[Callable[[Any], Any]] = None,
    **kwargs: Any,
) -> bytes:
    return JSONLIB.dumpb(
        data=data,
        pretty=pretty,
        sort_keys=sort_keys,
        append_newline=append_newline,
        default=default,
        **kwargs,
    )


def loads(string: str, **kwargs: Any) -> Any:
    return JSONLIB.loads(string, **kwargs)


def use_orjson() -> None:
    global JSONLIB
    JSONLIB = _import_orjson()


def use_rapidjson() -> None:
    global JSONLIB
    JSONLIB = _import_rapidjson()


def use_json_stdlib() -> None:
    global JSONLIB
    JSONLIB = _import_json_stdlib()


def use_json() -> None:
    use_json_stdlib()


def which() -> str:
    return JSONLIB.lib



# -*- coding: utf-8 -*-
import os

from typing import Any

from jsonbourne import JSON

try:
    from starlette.responses import Response
except ModuleNotFoundError:
    if not ('CI' in os.environ and os.environ['CI'] == 'true'):
        raise ModuleNotFoundError(
            'starlette not found/installed; `pip install starlette`'
        )
    else:
        Response = object  # type: ignore

__all__ = ('JSONBOURNEResponse',)


class JSONBOURNEResponse(Response):
    """FastAPI/starlette json response to auto use jsonbourne"""

    media_type = 'application/json'

    def render(self, content: Any) -> bytes:
        """Return JSON string for content as bytes"""
        return JSON.binify(data=content)



# -*- coding: utf-8 -*-
"""JSONBourne + Pydantic"""

from functools import lru_cache
from pprint import pformat
from typing import Any, Dict, Set, Type, TypeVar

# try:
from pydantic import BaseConfig, BaseModel, Extra, Field, ValidationError
from pydantic.generics import GenericModel

from jsonbourne.core import JSON, JsonObj

# except ImportError as ie:
#     raise ImportError('pydantic not found; install pydantid (`pip install pydantic`)')


JsonBaseModelT = TypeVar('JsonBaseModelT', bound='JsonBaseModel')

__all__ = (
    'JsonBaseModelDefaultConfig',
    'JsonBaseModel',
    'JsonBaseModelT',
    'BaseModel',
    'Field',
    'ValidationError',
    'JsonGenericModel',
    'GenericModel',
)


class JsonBaseModelDefaultConfig(BaseConfig):
    """Pydantic model config class for JsonBaseModel; can be overridden"""

    # Sometimes hypothesis breaks and will try to add an attribute to
    # objects while testing. Ya can check for 'pytest' if hypothesis breaks
    # (which it sometimes does) with `'pytest' in sys.modules`
    extra = Extra.forbid  # Forbid extras as strictness is good w/ python
    arbitrary_types_allowed = True  # Allow
    allow_population_by_field_name = True
    json_loads = JSON.loads
    json_dumps = JSON.dumps
    use_enum_values = True


class JsonBaseModel(BaseModel, JsonObj):  # type: ignore
    """Hybrid `pydantic.BaseModel` and `jsonbourne.JsonObj`"""

    Config = JsonBaseModelDefaultConfig

    def __init__(self, *args: Any, **kwargs: Any):  # type: ignore
        """Construct a JsonBaseModel and allow for `__post_init__` functions"""
        super().__init__(*args, **kwargs)
        self.__post_init__()

    def __post_init__(self) -> Any:
        """Function place holder that is called after object initialization"""
        pass  # pylint: disable=unnecessary-pass

    def to_str(
        self, minify: bool = False, width: int = 120, fmt_kwargs: bool = False
    ) -> str:
        if fmt_kwargs:
            return type(self).__name__ + '(' + self.__repr_str__(', ') + ')'
        if minify:
            return type(self).__name__ + '(**' + str(self.to_dict_filter_none()) + ')'
        return ''.join(
            [
                type(self).__name__,
                '(**{\n     ',
                pformat(self.to_dict_filter_none(), width=width)[1:-1].replace(
                    '\n', '\n    '
                ),
                '\n})',
            ]
        )

    def __repr__(self) -> str:
        return self.to_str(fmt_kwargs=True)

    def __str__(self) -> str:
        return self.to_str(minify=False)

    def _repr_html_(self) -> str:
        """Return the HTML representation of the object"""
        return '<pre>{}</pre>'.format(self.__str__())

    def to_dict_filter_none(self) -> Dict[str, Any]:
        """Eject object and filter key-values equal to (sub)class' default

        Examples:
            >>> from typing import Optional
            >>> class Thing(JsonBaseModel):
            ...     a: int = 1
            ...     b: str = "herm"
            ...     c: Optional[str] = None
            ...
            >>> t = Thing()
            >>> t
            Thing(a=1, b='herm', c=None)
            >>> t.to_dict_filter_none()
            {'a': 1, 'b': 'herm'}
            >>> t.to_json_obj_filter_none()
            JsonObj(**{'a': 1, 'b': 'herm'})

        """
        return {k: v for k, v in self.dict().items() if v is not None}

    def to_dict_filter_defaults(self) -> Dict[str, Any]:
        """Eject object and filter key-values equal to (sub)class' default

        Examples:
            >>> class Thing(JsonBaseModel):
            ...     a: int = 1
            ...     b: str = "herm"
            ...
            >>> t = Thing()
            >>> t
            Thing(a=1, b='herm')
            >>> t.to_dict_filter_defaults()
            {}
            >>> t.to_json_obj_filter_defaults()
            JsonObj(**{})
            >>> t = Thing(a=123)
            >>> t
            Thing(a=123, b='herm')
            >>> t.to_dict_filter_defaults()
            {'a': 123}
            >>> t.to_json_obj_filter_defaults()
            JsonObj(**{'a': 123})

        """
        defaults = self.defaults_dict()
        return {
            k: v
            for k, v in self.dict().items()
            if k not in defaults or v != defaults[k]
        }

    def to_json_obj_filter_defaults(self) -> JsonObj:
        """Eject to JsonObj and filter key-values equal to (sub)class' default"""
        return JsonObj(self.to_dict_filter_defaults())

    def to_json_obj_filter_none(self) -> JsonObj:
        """Eject to JsonObj and filter key-values where the value is None"""
        return JsonObj(self.to_dict_filter_none())

    def to_json_dict(self) -> JsonObj:
        """Eject object and sub-objects to `jsonbourne.JsonObj`

        Examples:
            >>> d = JsonObj(**{'uno': 'ONE', 'tres': 3, 'dos': 2})
            >>> d
            JsonObj(**{'uno': 'ONE', 'tres': 3, 'dos': 2})
            >>> plain_ol_dict = d.eject()
            >>> plain_ol_dict
            {'uno': 'ONE', 'tres': 3, 'dos': 2}
            >>> type(plain_ol_dict)
            <class 'dict'>

        """
        return JsonObj(
            {
                k: v if not isinstance(v, JsonBaseModel) else v.to_json_dict()
                for k, v in self.__dict__.items()
            }
        )

    def to_dict(self) -> Dict[str, Any]:
        """Eject and return object as plain jane dictionary"""
        return self.eject()

    @classmethod
    def from_dict_filtered(
        cls: Type[JsonBaseModelT], dictionary: Dict[str, Any]
    ) -> JsonBaseModelT:
        """Create class from dict filtering keys not in (sub)class' fields"""
        attr_names: Set[str] = set(cls._cls_field_names())
        return cls(**{k: v for k, v in dictionary.items() if k in attr_names})

    @classmethod
    def has_required_fields(cls) -> bool:
        """Return True/False if the (sub)class has any fields that are required

        Returns:
            bool: True if any fields for a (sub)class are required

        """
        return any(val.required for val in cls.__fields__.values())

    def is_default(self) -> bool:
        """Check if the object is equal to the default value for its fields

        Returns:
            True if object is equal to the default value for all fields; False otherwise

        Examples:
            >>> class Thing(JsonBaseModel):
            ...    a: int = 1
            ...    b: str = 'b'
            ...
            >>> t = Thing()
            >>> t.is_default()
            True
            >>> t = Thing(a=2)
            >>> t.is_default()
            False

        """
        if self.has_required_fields():
            return False
        return all(
            mfield.default == self[fname] for fname, mfield in self.__fields__.items()
        )

    def __delattr__(self, item: str) -> Any:
        if item in self.__private_attributes__:
            return object.__delattr__(self, item)
        return super().__delattr__(item)

    def __getattr__(self, item: str) -> Any:
        if item in self.__private_attributes__:
            return object.__getattribute__(self, item)
        return super().__getattr__(item)

    @classmethod
    def defaults_dict(cls) -> Dict[str, Any]:
        """Return a dictionary of non-required keys -> default value(s)

        Returns:
            Dict[str, Any]: Dictionary of non-required keys -> default value

        Examples:
            >>> class Thing(JsonBaseModel):
            ...     a: int = 1
            ...     b: str = "herm"
            ...
            >>> t = Thing()
            >>> t
            Thing(a=1, b='herm')
            >>> t.to_dict_filter_defaults()
            {}
            >>> t.to_json_obj_filter_defaults()
            JsonObj(**{})
            >>> t = Thing(a=123)
            >>> t
            Thing(a=123, b='herm')
            >>> t.to_dict_filter_defaults()
            {'a': 123}
            >>> t.to_json_obj_filter_defaults()
            JsonObj(**{'a': 123})
            >>> t.defaults_dict()
            {'a': 1, 'b': 'herm'}

        """
        return {k: v.default for k, v in cls.__fields__.items() if not v.required}

    def __setattr__(self, name: str, value: Any) -> None:
        if name in self.__private_attributes__:
            return object.__setattr__(self, name, value)
        elif name in self.__property_fields__:
            property_field = getattr(self.__class__, name)
            property_field.fset(self, value)
        else:
            super().__setattr__(name, value)

    @property
    def __property_fields__(self) -> Set[str]:
        """Returns a set of property names for the class that have a setter"""
        return self.__class__._cls_property_fields()

    @classmethod
    @lru_cache(maxsize=1)
    def _cls_property_fields(cls) -> Set[str]:
        """Return a set of property names with a setter function"""
        return {
            k
            for k, v in ((el, getattr(cls, el)) for el in dir(cls))
            if isinstance(v, property) and v.fset is not None
        }

    @classmethod
    def _cls_field_names(cls) -> Set[str]:
        """Return pydantic field names"""
        return set(cls.__fields__)

    def _field_names(self) -> Set[str]:
        """Return pydantic field names"""
        return self.__class__._cls_field_names()


class JsonGenericModel(GenericModel, JsonBaseModel):
    """Hybrid `pydantic.generics.GenericModel` and `jsonbourne.JsonObj`"""

    ...



# -*- coding: utf-8 -*-
"""Json Bourne -- EZ-PZ-JSON with lots o goodies"""
import keyword
import sys

from functools import lru_cache
from itertools import chain
from json import JSONDecodeError
from pprint import pformat
from types import ModuleType
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    ItemsView,
    Iterable,
    Iterator,
    KeysView,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    overload,
)

from jsonbourne import jsonlib

JsonPrimitiveT = TypeVar('JsonPrimitiveT', str, int, float, None)
JsonObjT = TypeVar('JsonObjT', bound='JsonObj')
KT = TypeVar('KT')
VT = TypeVar('VT')
_KT = str
_VT = TypeVar('_VT')

if sys.version_info < (3, 7):
    from collections.abc import MutableMapping

    JsonObjMutableMapping = MutableMapping
else:
    from typing import MutableMapping

    JsonObjMutableMapping = MutableMapping[str, _VT]

__all__ = (
    'JsonObj',
    'JsonDict',
    'JsonObjMutableMapping',
    'stringify',
    'parse',
    'jsonify',
    'JSON',
    'UNDEFINED',
    'undefined',
    'Null',
    'null',
    'JSONModuleCls',
    'JsonObjT',
)

_JsonObjMutableMapping_attrs = set(dir(JsonObjMutableMapping))

UNDEFINED = 'undefined'
undefined = 'undefined'
null = None
Null = None.__class__


def is_identifier(string: str) -> bool:
    """Return True if a string is a valid python identifier; False otherwise

    Args:
        string (str): String (likely to be used as a key) to check

    Returns:
        bool: True if is an identifier

    Examples:
        >>> is_identifier("herm")
        True
        >>> is_identifier("something that contains spaces")
        False
        >>> is_identifier("import")
        False
        >>> is_identifier("something.with.periods")
        False
        >>> is_identifier("astring-with-dashes")
        False
        >>> is_identifier("astring_with_underscores")
        True
        >>> is_identifier(123)
        False

    """
    try:
        assert isinstance(string, str)
    except AssertionError:
        return False
    if not string.isidentifier():
        return False
    if keyword.iskeyword(string):
        return False
    return True


def is_float(value: Any) -> bool:
    """Return True if value is a float"""
    try:
        float(value)
        return True
    except ValueError:
        pass
    return False


def is_int(value: Any) -> bool:
    """Return True if value is a int"""
    if isinstance(value, int):
        return True
    _value: str = str(value)
    if _value[0] in ('-', '+'):
        return str(_value[1:]).isdigit()
    return _value.isdigit()


def is_number(value: Any) -> bool:
    """Return True if value is a number"""
    return is_int(value) or is_float(value)


class JsonObj(JsonObjMutableMapping, Generic[_VT]):
    """JSON friendly python dictionary with dot notation and string only keys

    JsonObj(foo='bar')['foo'] == JsonObj(foo='bar').foo

    Examples:
        >>> d = {"uno": 1, "dos": 2, "tres": 3}
        >>> d
        {'uno': 1, 'dos': 2, 'tres': 3}
        >>> d = JsonObj(d)
        >>> d
        JsonObj(**{'uno': 1, 'dos': 2, 'tres': 3})
        >>> list(d.keys())
        ['uno', 'dos', 'tres']
        >>> list(d.dot_keys())
        [('uno',), ('dos',), ('tres',)]
        >>> d
        JsonObj(**{'uno': 1, 'dos': 2, 'tres': 3})
        >>> d['uno']
        1
        >>> d.uno
        1
        >>> d['uno'] == d.uno
        True
        >>> d.uno = "ONE"
        >>> d
        JsonObj(**{'uno': 'ONE', 'dos': 2, 'tres': 3})
        >>> d['uno'] == d.uno
        True
        >>> 'uno' in d
        True
        >>> 'not_in_d' in d
        False
        >>> d
        JsonObj(**{'uno': 'ONE', 'dos': 2, 'tres': 3})
        >>> del d['dos']
        >>> d
        JsonObj(**{'uno': 'ONE', 'tres': 3})
        >>> d.tres
        3
        >>> del d.tres
        >>> d
        JsonObj(**{'uno': 'ONE'})
        >>> d = {"uno": 1, "dos": 2, "tres": {"a": 1, "b": [3, 4, 5, 6]}}
        >>> d = JsonObj(d)
        >>> d
        JsonObj(**{'uno': 1, 'dos': 2, 'tres': {'a': 1, 'b': [3, 4, 5, 6]}})
        >>> d.tres
        JsonObj(**{'a': 1, 'b': [3, 4, 5, 6]})
        >>> d.tres.a
        1
        >>> d.tres.a = "new-val"
        >>> d.tres.a
        'new-val'
        >>> d
        JsonObj(**{'uno': 1, 'dos': 2, 'tres': {'a': 'new-val', 'b': [3, 4, 5, 6]}})
        >>> jd = JsonObj({"a":1, "b": 'herm', 'alist':[{'sub': 123}]})

        It does lists!? oh my

        >>> jd
        JsonObj(**{'a': 1, 'b': 'herm', 'alist': [{'sub': 123}]})
        >>> jd.alist[0]
        JsonObj(**{'sub': 123})
        >>> jd.eject()
        {'a': 1, 'b': 'herm', 'alist': [{'sub': 123}]}

    """

    _data: Dict[_KT, _VT]

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Use the object dict"""
        _data = dict(*args, **kwargs)
        super().__setattr__('_data', _data)
        try:
            assert all(isinstance(k, str) for k in self._data)
        except AssertionError:
            d = {k: v for k, v in self._data.items() if not isinstance(k, str)}
            raise ValueError(
                'JsonObj keys MUST be strings! Bad key values: {}'.format(str(d))
            )
        self.recurse()
        self.__post_init__()

    def recurse(self) -> None:
        """Recusively convert all sub dictionaries to JsonObj objects"""
        self._data.update({k: jsonify(v) for k, v in self._data.items()})

    def __attrs_post_init__(self) -> None:
        self.recurse()

    def __post_init__(self) -> Any:
        """Function place holder that is called after object initialization"""
        pass  # pylint: disable=unnecessary-pass

    def __contains__(self, key: _KT) -> bool:  # type: ignore
        """Check if a key or dot-key is contained within the JsonObj object

        Args:
            key (str): root level key or a dot-key

        Returns:
            bool: True if the key/dot-key is in the JsonObj; False otherwise

        Examples:
            >>> d = {"uno": 1, "dos": 2, "tres": 3, "sub": {"a": 1, "b": 2, "c": [3, 4, 5, 6], "d": "a_string"}}
            >>> d = JsonObj(d)
            >>> d
            JsonObj(**{'uno': 1, 'dos': 2, 'tres': 3, 'sub': {'a': 1, 'b': 2, 'c': [3, 4, 5, 6], 'd': 'a_string'}})
            >>> 'uno' in d
            True
            >>> 'this_key_is_not_in_d' in d
            False
            >>> 'sub.a' in d
            True
            >>> 'sub.d.a' in d
            False

        """
        if '.' in key:
            first_key, _, rest = key.partition('.')
            val = self._data.get(first_key)
            return isinstance(val, MutableMapping) and val.__contains__(rest)
        return key in self._data

    def __setattr__(self, attr: _KT, value: _VT) -> None:
        if attr in self._cls_protected_attrs():
            raise ValueError(
                f"Cannot set protected attribute ('{str(attr)}'),"
                f" must use brackets/setitem syntax: json_obj['{str(attr)}']"
            )
        return self.__setitem__(attr, value)

    def __setitem__(self, key: _KT, value: _VT) -> None:
        """Set JsonObj item with 'key' to 'value'

        Args:
            key (str): Key/item to set
            value: Value to set

        Returns:
            None

        Raises:
            ValueError: If given a key that is not a valid python keyword/identifier

        Examples:
            >>> d = JsonObj()
            >>> d.a = 123
            >>> d['b'] = 321
            >>> d
            JsonObj(**{'a': 123, 'b': 321})
            >>> d[123] = 'a'
            >>> d
            JsonObj(**{'a': 123, 'b': 321, '123': 'a'})
            >>> d['456'] = 'a'
            >>> d
            JsonObj(**{'a': 123, 'b': 321, '123': 'a', '456': 'a'})

        """
        if is_int(key):
            self._data[str(key)] = value
            return None
        if not is_identifier(key):
            raise ValueError(
                f'Invalid key: ({key}).\n' f'Key(s) is not a valid python identifier'
            )
        self._data[key] = value

    def __getattr__(self, item: _KT) -> Any:
        """Return an attr

        Examples:
            >>> d = {
            ...     'falsey_dict': {},
            ...     'falsey_list': [],
            ...     'falsey_string': '',
            ...     'is_false': False,
            ...     'a': None,
            ...     'b': 2,
            ...     'c': {
            ...         'd': 'herm',
            ...         'e': None,
            ...         'falsey_dict': {},
            ...         'falsey_list': [],
            ...         'falsey_string': '',
            ...         'is_false': False,
            ...     },
            ...     }
            ...
            >>> d = JsonObj(d)
            >>> d.__getattr__('b')
            2
            >>> d.b
            2

        """
        if item == '_data':
            try:
                return object.__getattribute__(self, '_data')
            except AttributeError:
                return self.__dict__
        if item in _JsonObjMutableMapping_attrs or item in self._cls_protected_attrs():
            return object.__getattribute__(self, item)
        try:
            return jsonify(self.__getitem__(str(item)))
        except KeyError:
            pass
        return object.__getattribute__(self, item)

    def __object_getattribute__(self, item: str) -> Any:
        return object.__getattribute__(self, item)

    def __getitem__(self, key: Union[_KT, Tuple[_KT, ...]]) -> Any:
        if isinstance(key, str):
            try:
                return jsonify(self._data[key])
            except KeyError:
                pass
            try:
                return jsonify(self.__object_getattribute__(key))
            except AttributeError:
                pass
        try:
            return self.dot_lookup(key)
        except KeyError:
            ...
        raise KeyError(str(key))

    def __delitem__(self, key: _KT) -> None:
        return self._data.__delitem__(key)

    def __delattr__(self, item: _KT) -> None:
        return self.__delitem__(item)

    def __iter__(self) -> Iterator[_KT]:
        return iter(self._data)

    def __len__(self) -> int:
        return len(self._data)

    def items(self) -> ItemsView[_KT, _VT]:
        """Return an items view of the JsonObj object"""
        return self._data.items()

    def entries(self) -> ItemsView[_KT, _VT]:
        """Alias for items"""
        return self.items()

    def keys(self) -> KeysView[_KT]:
        """Return the keys view of the JsonObj object"""
        return self._data.keys()

    def setdefault(self, key: _KT, default: Optional[_VT] = None) -> _VT:
        if default:
            return self._data.setdefault(key, default)
        return self._data.setdefault(key)

    def clear(self) -> None:
        self._data.clear()

    def pop(self, key: _KT, default: Optional[Any] = None) -> Any:
        if default:
            return self._data.pop(key, default)
        return self._data.pop(key)

    def get(self, key: _KT, default: Optional[Any] = None) -> Any:
        try:
            return self._data.get(key)
        except KeyError as ke:
            if default:
                return default
            raise ke

    def filter_none(self, recursive: bool = False) -> 'JsonObj':
        """Filter key-values where the value is `None` but not false-y

        Args:
            recursive (bool): Recursively filter out None values

        Returns:
            JsonObj that has been filtered of None values

        Examples:
            >>> d = {
            ...     'falsey_dict': {},
            ...     'falsey_list': [],
            ...     'falsey_string': '',
            ...     'is_false': False,
            ...     'a': None,
            ...     'b': 2,
            ...     'c': {
            ...         'd': 'herm',
            ...         'e': None,
            ...         'falsey_dict': {},
            ...         'falsey_list': [],
            ...         'falsey_string': '',
            ...         'is_false': False,
            ...     },
            ...     }
            ...
            >>> d = JsonObj(d)
            >>> print(d)
            JsonObj(**{
                'a': None,
                'b': 2,
                'c': {'d': 'herm',
                      'e': None,
                      'falsey_dict': {},
                      'falsey_list': [],
                      'falsey_string': '',
                      'is_false': False},
                'falsey_dict': {},
                'falsey_list': [],
                'falsey_string': '',
                'is_false': False
            })
            >>> print(d.filter_none())
            JsonObj(**{
                'b': 2,
                'c': {'d': 'herm',
                      'e': None,
                      'falsey_dict': {},
                      'falsey_list': [],
                      'falsey_string': '',
                      'is_false': False},
                'falsey_dict': {},
                'falsey_list': [],
                'falsey_string': '',
                'is_false': False
            })
            >>> from pprint import pprint
            >>> print(d.filter_none(recursive=True))
            JsonObj(**{
                'b': 2,
                'c': {'d': 'herm',
                      'falsey_dict': {},
                      'falsey_list': [],
                      'falsey_string': '',
                      'is_false': False},
                'falsey_dict': {},
                'falsey_list': [],
                'falsey_string': '',
                'is_false': False
            })

        """
        if recursive:
            return JsonObj(
                {
                    k: v
                    if not isinstance(v, (dict, JsonObj))
                    else JsonObj(v).filter_none(recursive=True)
                    for k, v in self.items()
                    if v is not None
                }
            )
        return JsonObj({k: v for k, v in self.items() if v is not None})

    def filter_false(self, recursive: bool = False) -> 'JsonObj':
        """Filter key-values where the value is false-y

        Args:
            recursive (bool): Recurse into sub JsonObjs and dictionaries

        Returns:
            JsonObj that has been filtered

        Examples:
            >>> d = {
            ...     'falsey_dict': {},
            ...     'falsey_list': [],
            ...     'falsey_string': '',
            ...     'is_false': False,
            ...     'a': None,
            ...     'b': 2,
            ...     'c': {
            ...         'd': 'herm',
            ...         'e': None,
            ...         'falsey_dict': {},
            ...         'falsey_list': [],
            ...         'falsey_string': '',
            ...         'is_false': False,
            ...     },
            ...     }
            ...
            >>> d = JsonObj(d)
            >>> print(d)
            JsonObj(**{
                'a': None,
                'b': 2,
                'c': {'d': 'herm',
                      'e': None,
                      'falsey_dict': {},
                      'falsey_list': [],
                      'falsey_string': '',
                      'is_false': False},
                'falsey_dict': {},
                'falsey_list': [],
                'falsey_string': '',
                'is_false': False
            })
            >>> print(d.filter_false())
            JsonObj(**{
                'b': 2,
                'c': {'d': 'herm',
                      'e': None,
                      'falsey_dict': {},
                      'falsey_list': [],
                      'falsey_string': '',
                      'is_false': False}
            })
            >>> print(d.filter_false(recursive=True))
            JsonObj(**{
                'b': 2, 'c': {'d': 'herm'}
            })

        """
        if recursive:
            return JsonObj(
                {
                    k: v
                    if not isinstance(v, (dict, JsonObj))
                    else JsonObj(v).filter_false(recursive=True)
                    for k, v in self.items()
                    if v
                }
            )
        return JsonObj({k: v for k, v in self.items() if v})

    def dot_keys(self) -> Iterable[Tuple[str, ...]]:
        """Yield the JsonObj's dot-notation keys

        Returns:
            Iterable[str]: List of the dot-notation friendly keys


        The Non-chain version (shown below) is very slightly slower than the
        `itertools.chain` version.

        NON-CHAIN VERSION:

        for k, value in self.items():
            value = jsonify(value)
            if isinstance(value, JsonObj):
                yield from (f"{k}.{dk}" for dk in value.dot_keys())
            else:
                yield k

        """
        return chain(
            *(
                ((str(k),),)
                if not isinstance(v, JsonObj)
                else (*((str(k), *dk) for dk in jsonify(v).dot_keys()),)
                for k, v in self.items()
            )
        )

    def dot_keys_list(self, sort_keys: bool = False) -> List[Tuple[str, ...]]:
        """Return a list of the JsonObj's dot-notation friendly keys

        Args:
            sort_keys (bool): Flag to have the dot-keys be returned sorted

        Returns:
            List[str]: List of the dot-notation friendly keys

        """
        if sort_keys:
            return sorted(self.dot_keys_list())
        return list(self.dot_keys())

    def dot_keys_set(self) -> Set[Tuple[str, ...]]:
        """Return a set of the JsonObj's dot-notation friendly keys

        Returns:
            Set[str]: List of the dot-notation friendly keys

        """
        return set(self.dot_keys())

    def dot_lookup(self, key: Union[str, Tuple[str, ...], List[str]]) -> Any:
        """Look up JsonObj keys using dot notation as a string

        Args:
            key (str): dot-notation key to look up ('key1.key2.third_key')

        Returns:
            The result of the dot-notation key look up

        Raises:
            KeyError: Raised if the dot-key is not in in the object

        """
        if not isinstance(key, (str, list, tuple)):
            raise ValueError(
                ''.join(
                    (
                        'dot_key arg must be string or sequence of strings; ',
                        "strings will be split on '.'",
                    )
                )
            )
        parts = key.split('.') if isinstance(key, str) else list(key)
        root_val: Any = self._data[parts[0]]
        cur_val = root_val
        for ix, part in enumerate(parts[1:], start=1):
            try:
                cur_val = cur_val[part]
            except TypeError:
                reached = '.'.join(parts[:ix])
                err_msg = f'Invalid DotKey: {key} -- Lookup reached: {reached} => {str(cur_val)}'
                if isinstance(key, str):
                    err_msg += ''.join(
                        (
                            f"\nNOTE!!! lookup performed with string ('{key}') ",
                            'PREFER lookup using List[str] or Tuple[str, ...]',
                        )
                    )
                raise KeyError(err_msg)
        return cur_val

    def dot_items(self) -> Iterator[Tuple[Tuple[str, ...], _VT]]:
        """Yield tuples of the form (dot-key, value)

        OG-version:
            def dot_items(self) -> Iterator[Tuple[str, Any]]:
                return ((dk, self.dot_lookup(dk)) for dk in self.dot_keys())

        Readable-version:
            for k, value in self.items():
                value = jsonify(value)
                if isinstance(value, JsonObj) or hasattr(value, 'dot_items'):
                    yield from ((f"{k}.{dk}", dv) for dk, dv in value.dot_items())
                else:
                    yield k, value
        """

        return chain.from_iterable(
            (
                (
                    *(
                        (
                            (
                                str(k),
                                *dk,
                            ),
                            dv,
                        )
                        for dk, dv in as_json_obj(v).dot_items()
                    ),
                )
                if isinstance(v, (JsonObj, dict))
                else (((str(k),), v),)
                for k, v in self.items()
            )
        )

    def dot_items_list(self) -> List[Tuple[Tuple[str, ...], Any]]:
        """Return list of tuples of the form (dot-key, value)"""
        return list(self.dot_items())

    def to_str(self, minify: bool = False, width: int = 88) -> str:
        """Return a string representation of the JsonObj object"""
        if minify:
            return type(self).__name__ + '(**' + str(self.to_dict()) + ')'
        return ''.join(
            [
                type(self).__name__,
                '(**{\n    ',
                pformat(self.to_dict(), width=width)[1:-1].replace('\n', '\n   '),
                '\n})',
            ]
        ).replace('JsonObj(**{}),', '{},')

    def __repr__(self) -> str:
        """Return the string representation of the object"""
        return self.to_str(minify=True)

    def __str__(self) -> str:
        """Return the string representation of the JsonObj object"""
        return self.to_str(minify=False)

    def _repr_html_(self) -> str:
        """Return the HTML representation of the JsonObj object"""
        return '<pre>{}</pre>'.format(self.__str__())

    @classmethod
    def _cls_attr_names(cls) -> Set[str]:
        """Return attrs-attribute names for an object decorated with attrs"""
        try:
            return {el.name for el in cls.__attrs_attrs__}  # type: ignore
        except AttributeError:
            raise AttributeError('Class is not decorated with attr.attrs')

    @classmethod
    def _cls_fields(cls) -> Set[str]:
        """Return attrs-attribute names for an object decorated with attrs"""
        try:
            return cls.__fields__  # type: ignore
        except AttributeError:
            raise AttributeError('Class does not inherit from pydantic.BaseModel')

    @classmethod
    def _cls_field_names(cls) -> Set[str]:
        """Return attrs-attribute names for an object decorated with attrs"""
        try:
            return set(cls.__fields__)  # type: ignore
        except AttributeError:
            raise AttributeError('Class does not inherit from pydantic.BaseModel')

    @classmethod
    def _cls_protected_attrs(cls) -> Set[str]:
        """Return attrs-attribute names for an object decorated with attrs"""
        return {
            'asdict',
            'clear',
            'copy',
            'dot_items',
            'dot_items_list',
            'dot_keys',
            'dot_keys_list',
            'dot_keys_set',
            'dot_lookup',
            'eject',
            'entries',
            'filter_false',
            'filter_none',
            'from_dict',
            'from_json',
            'fromkeys',
            'get',
            'items',
            'keys',
            'pop',
            'popitem',
            'recurse',
            'setdefault',
            'stringify',
            'to_dict',
            'to_json',
            'to_str',
            'update',
            'validate_type',
            'values',
        }

    def _field_names(self) -> Set[str]:
        """Return attrs-attribute names for an object decorated with attrs"""
        return self.__class__._cls_field_names()

    def eject(self) -> Dict[_KT, Any]:
        """Eject to python-builtin dictionary object

        Examples:
            >>> d = JsonObj(**{'uno': 'ONE', 'tres': 3, 'dos': 2})
            >>> d
            JsonObj(**{'uno': 'ONE', 'tres': 3, 'dos': 2})
            >>> plain_ol_dict = d.eject()
            >>> plain_ol_dict
            {'uno': 'ONE', 'tres': 3, 'dos': 2}
            >>> type(plain_ol_dict)
            <class 'dict'>

        """
        try:
            return {k: unjsonify(v) for k, v in self._data.items()}
        except RecursionError:
            raise ValueError(
                'JSON.stringify recursion err; cycle/circular-refs detected'
            )

    def to_dict(self) -> Dict[_KT, Any]:
        """Return the JsonObj object (and children) as a python dictionary"""
        return self.eject()

    def asdict(self) -> Dict[_KT, Any]:
        """Return the JsonObj object (and children) as a python dictionary"""
        return self.eject()

    @classmethod
    def from_dict(cls: Type[JsonObjT], data: Dict[_KT, _VT]) -> JsonObjT:
        """Return a JsonObj object from a dictionary of data"""
        return cls(**data)

    @classmethod
    def from_json(cls: Type[JsonObjT], json_string: str) -> JsonObjT:
        """Return a JsonObj object from a json string

        Args:
            json_string (str): JSON string to convert to a JsonObj

        Returns:
            JsonObj: JsonObj object for the given JSON string

        """
        return cls._from_json(json_string)

    @classmethod
    def _from_json(cls: Type[JsonObjT], json_string: str) -> JsonObjT:
        """Return a JsonObj object from a json string

        Args:
            json_string (str): JSON string to convert to a JsonObj

        Returns:
            JsonObj: JsonObj object for the given JSON string

        """
        return cls.from_dict(jsonlib.loads(json_string))

    def to_json(
        self, pretty: bool = False, sort_keys: bool = False, **kwargs: Any
    ) -> str:
        """Return a JSON string of the JsonObj object

        Args:
            minify (bool): Return a 'minified' version of the JSON string
            sort_keys (bool): Sort the keys when converting to JSON
            indent (int): Indent level of the json string
            **kwargs: Keyword args to be passed on to the JSON dumps method

        Returns:
            str: JSON string of the JsonObj object

        """
        return self._to_json(pretty=pretty, sort_keys=sort_keys, **kwargs)

    def stringify(
        self, pretty: bool = False, sort_keys: bool = False, **kwargs: Any
    ) -> str:
        """Return a JSON string of the JsonObj; `JsonObj.to_json` alias

        Args:
            pretty (bool): If a 'minified' version of the JSON string
            sort_keys (bool): Sort the keys when converting to JSON
            **kwargs: Keyword args to be passed on to the JSON dumps method

        Returns:
            str: JSON string of the JsonObj object

        """
        return self.to_json(pretty=pretty, sort_keys=sort_keys, **kwargs)

    def _to_json(
        self, pretty: bool = False, sort_keys: bool = False, **kwargs: Any
    ) -> str:
        """Return a JSON string of the JsonObj object

        Args:
            minify (bool): Return a 'minified' version of the JSON string
            sort_keys (bool): Sort the keys when converting to JSON
            indent (int): Indent level of the json string
            **kwargs: Keyword args to be passed on to the JSON dumps method

        Returns:
            str: JSON string of the JsonObj object

        """
        return jsonlib.dumps(
            self.to_dict(), pretty=pretty, sort_keys=sort_keys, **kwargs
        )

    @classmethod
    def validate_type(cls: Type[JsonObjT], val: Any) -> JsonObjT:
        """Validate and convert a value to a JsonObj object"""
        return cls(val)

    @classmethod
    def __get_validators__(cls) -> Iterator[Callable[[Any], Any]]:
        """Return the JsonObj validator functions"""
        yield cls.validate_type


class JsonDict(JsonObj):
    """Alias for JsonObj"""

    pass


# @overload
# def jsonify(value: Dict[_KT, _VT]) -> JsonObj[_VT]:
#     ...


def as_json_obj(value: Union[JsonObj[_VT], Dict[_KT, _VT]]) -> JsonObj[_VT]:
    if isinstance(value, dict):
        return JsonObj(value)
    return value


@overload
def jsonify(value: JsonPrimitiveT) -> JsonPrimitiveT:
    ...


@overload
def jsonify(value: List[JsonPrimitiveT]) -> List[JsonPrimitiveT]:
    ...


@overload
def jsonify(value: Tuple[JsonPrimitiveT, ...]) -> Tuple[JsonPrimitiveT, ...]:
    ...


@overload
def jsonify(value: _VT) -> _VT:
    ...


def jsonify(value: Any) -> Any:
    """Convert and return a value to a JsonObj if the value is a dict"""
    if isinstance(value, (JsonObj, JsonDict)) or issubclass(value.__class__, JsonObj):
        return value
    if isinstance(value, dict) and not isinstance(value, JsonObj):
        return JsonObj(value)
    if isinstance(value, list):
        return [jsonify(el) for el in value]
    if isinstance(value, tuple):
        return tuple([jsonify(el) for el in value])
    if isinstance(value, str):
        try:
            data = jsonlib.loads(value)
            return jsonify(data)
        except Exception:
            pass

    return value


def unjsonify(value: Any) -> Any:
    """Recursively eject a JsonDit object"""
    if isinstance(value, JsonObj):
        return value._data
    if isinstance(value, list):
        return [unjsonify(el) for el in value]
    if isinstance(value, tuple):
        return tuple([unjsonify(el) for el in value])
    return value


class JSONMeta(type):
    """Meta type for use by JSON class to allow for static `__call__` method"""

    @staticmethod
    def __call__(value: Optional[Any] = None) -> Any:  # type: ignore
        if value is None:
            value = {}
        return jsonify(value)


class JSON(metaclass=JSONMeta):
    """JSON class meant to mimic the js/ts-JSON"""

    undefined: str = UNDEFINED
    UNDEFINED: str = UNDEFINED
    null = null
    Null = Null
    JSONDecodeError = JSONDecodeError

    @staticmethod
    def stringify(
        data: Any,
        pretty: bool = False,
        sort_keys: bool = False,
        default: Optional[Callable[[Any], Any]] = None,
        **kwargs: Any,
    ) -> str:
        """Return JSON stringified/dumps-ed data"""
        return str(
            jsonlib.dumps(
                data,
                pretty=pretty,
                sort_keys=sort_keys,
                default=default,
                **kwargs,
            )
        )

    @staticmethod
    def dumps(
        data: Any,
        pretty: bool = False,
        sort_keys: bool = False,
        default: Optional[Callable[[Any], Any]] = None,
        **kwargs: Any,
    ) -> str:
        """Return JSON stringified/dumps-ed data"""
        return str(
            jsonlib.dumps(
                data,
                pretty=pretty,
                sort_keys=sort_keys,
                default=default,
                **kwargs,
            )
        )

    @staticmethod
    def binify(
        data: Any,
        pretty: bool = False,
        sort_keys: bool = False,
        default: Optional[Callable[[Any], Any]] = None,
        **kwargs: Any,
    ) -> bytes:
        """Return JSON string bytes for given data"""
        return bytes(
            jsonlib.dumpb(
                data,
                pretty=pretty,
                sort_keys=sort_keys,
                default=default,
                **kwargs,
            )
        )

    @staticmethod
    def dumpb(
        data: Any,
        pretty: bool = False,
        sort_keys: bool = False,
        default: Optional[Callable[[Any], Any]] = None,
        **kwargs: Any,
    ) -> bytes:
        """Return JSON string bytes for given data"""
        return bytes(
            jsonlib.dumpb(
                data,
                pretty=pretty,
                sort_keys=sort_keys,
                default=default,
                **kwargs,
            )
        )

    @staticmethod
    def loads(string: str, obj: bool = False, **kwargs: Any) -> Any:
        """Parse JSON string/bytes and return raw representation"""
        if obj:
            return jsonify(jsonlib.loads(string, **kwargs))
        return jsonlib.loads(string, **kwargs)

    @staticmethod
    def parse(string: str, obj: bool = True) -> Any:
        """Parse JSON string/bytes and jsonify all dictionaries -> JsonObj"""
        if obj:
            return jsonify(jsonlib.loads(string))
        return jsonlib.loads(string)

    @staticmethod
    def use_orjson() -> None:
        jsonlib.use_orjson()

    @staticmethod
    def use_rapidjson() -> None:
        jsonlib.use_rapidjson()

    @staticmethod
    def use_json_stdlib() -> None:
        jsonlib.use_json_stdlib()

    @staticmethod
    def which() -> str:
        """Return the name of the JSON library being used as a backend"""
        return jsonlib.which()

    @staticmethod
    def json_lib() -> str:
        """Return the name of the JSON library being used as a backend"""
        return jsonlib.which()

    @staticmethod
    def jsonify(value: Any) -> Any:
        """Alias for jsonbourne.core.jsonify"""
        return jsonify(value)

    @staticmethod
    def unjsonify(value: Any) -> Any:
        """Alias for jsonbourne.core.unjsonify"""
        return unjsonify(value)


class JSONModuleCls(ModuleType, JSON):
    @staticmethod
    def __call__(value: Any = None):  # type: ignore
        """Jsonify a value"""
        if value is None:
            return JsonObj()
        return jsonify(value)


@lru_cache(maxsize=None)
def _cls_protected_attrs(cls: Any) -> Set[str]:
    """Return attrs-attribute names for an object decorated with attrs"""
    return set(dir(cls))


stringify = JSON.stringify
parse = JSON.parse

if __name__ == '__main__':
    import doctest

    doctest.testmod()



# -*- coding: utf-8 -*-
"""dataclasses ~ `pydantic.dataclasses.dataclass` > `dataclasses.dataclass`"""
from dataclasses import (
    MISSING,
    Field,
    FrozenInstanceError,
    InitVar,
    asdict,
    astuple,
    field,
    fields,
    is_dataclass,
    make_dataclass,
    replace,
)

try:
    # use `pydantic.dataclasses.dataclass` if available
    from pydantic.dataclasses import dataclass
except ImportError:
    from dataclasses import dataclass


__all__ = (
    'Field',
    'FrozenInstanceError',
    'InitVar',
    'MISSING',
    'asdict',
    'astuple',
    'dataclass',
    'field',
    'fields',
    'is_dataclass',
    'make_dataclass',
    'replace',
)



# -*- coding: utf-8 -*-
"""Json Bourne types"""
import sys

from typing import Any, Dict, List

if sys.version_info < (3, 7):
    from collections.abc import MutableMapping

    _JsonDictMutableMapping = MutableMapping
else:
    from typing import MutableMapping

    _JsonDictMutableMapping = MutableMapping[str, Any]

__all__ = (
    'JsonDictT',
    'JsonListT',
    'JsonArrT',
    'JsonObjT',
)

# Json friendly: Union[None, bool, int, float, str, List[Any], Dict[str, Any]]
JsonDictT = Dict[str, Any]
JsonObjT = Dict[str, Any]
JsonListT = List[Any]
JsonArrT = List[Any]



# -*- coding: utf-8 -*-
"""JSON callable module ~ jsonbourne"""
import sys

from jsonbourne import __version__, json
from jsonbourne.core import (
    JSON,
    UNDEFINED,
    JsonDict,
    JSONModuleCls,
    JsonObj,
    JsonObjMutableMapping,
    Null,
    null,
    undefined,
)
from jsonbourne.jsonlib import import_json

sys.modules['JSON'].__class__ = JSONModuleCls
stringify = JSON.stringify
dumps = JSON.dumps
binify = JSON.binify
dumpb = JSON.dumpb
loads = JSON.loads
parse = JSON.parse
json_lib = JSON.json_lib
which = JSON.which
use_orjson = JSON.use_orjson
use_rapidjson = JSON.use_rapidjson
use_json_stdlib = JSON.use_json_stdlib
jsonify = JSON.jsonify
unjsonify = JSON.unjsonify
JSONDecodeError = JSON.JSONDecodeError

__all__ = [
    '__version__',
    'json',  # json compat lib
    # core
    'JsonObjMutableMapping',
    'JsonObj',
    'JsonDict',
    'JSONModuleCls',
    # import
    'import_json',
    'use_orjson',
    'use_rapidjson',
    'use_json_stdlib',
    # util funks
    'stringify',
    'binify',
    'dumps',
    'dumpb',
    'loads',
    'json_lib',
    'jsonify',
    'unjsonify',
    'parse',
    'undefined',
    'UNDEFINED',
    'Null',
    'null',
]



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': '.'}

packages = \
['jsonbourne']

package_data = \
{'': ['*']}

modules = \
['JSON', 'david_webb']
extras_require = \
{'all': ['pydantic>=1.8.2', 'python-rapidjson>=0.9.1', 'orjson>=3.0.0,<4.0.0'],
 'full': ['pydantic>=1.8.2', 'python-rapidjson>=0.9.1', 'orjson>=3.0.0,<4.0.0'],
 'orjson': ['orjson>=3.0.0,<4.0.0'],
 'pydantic': ['pydantic>=1.8.2'],
 'rapidjson': ['python-rapidjson>=0.9.1']}

setup_kwargs = {
    'name': 'jsonbourne',
    'version': '0.18.0',
    'description': 'EZPZ JSON',
    'long_description': None,
    'author': 'jesse',
    'author_email': 'jesse@dgi.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/dynamic-graphics-inc/dgpy-libs',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'py_modules': modules,
    'extras_require': extras_require,
    'python_requires': '>=3.6.2,<4.0.0',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-
"""jsonbourne the best undercover json lib

Dynamic Graphics Python
"""
from jsonbourne import jsonlib
from jsonbourne._meta import __version__
from jsonbourne.core import (
    JSON,
    UNDEFINED,
    JsonDict,
    JsonObj,
    JsonObjMutableMapping,
    parse,
    stringify,
    undefined,
)
from jsonbourne.helpers import rm_js_comments
from jsonbourne.jsonlib import import_json

json = jsonlib

JSONLIB = jsonlib.__name__

__all__ = (
    '__version__',
    'JSON',  # js/ts JSON (THE ONE TO USE)
    'jsonlib',  # json compat lib
    'json',  # json compat lib
    # core
    'JsonObjMutableMapping',
    'JsonObj',
    'JsonDict',
    # import
    'import_json',
    # util funks
    'stringify',
    'parse',
    'rm_js_comments',
    # undefined
    'undefined',
    'UNDEFINED',
)
